import itertools as itr
from causaldag.utils.ci_tests import CI_Tester


def estimate_adjacent(nodes: set, target_node, ci_tester: CI_Tester):
    nnodes = len(nodes)
    nbrs = nodes - {target_node}
    for c_size in range(nnodes-1):
        nbrs_ = nbrs.copy()
        for nbr in nbrs_:
            if len(nbrs_-{nbr}) >= c_size:
                for cond_set in itr.combinations(nbrs_-{nbr}, c_size):
                    if ci_tester.is_ci(target_node, nbr, cond_set):
                        nbrs.remove(nbr)
                        break
    return nbrs


if __name__ == '__main__':
    import causaldag as cd
    from causaldag.utils.ci_tests import MemoizedCI_Tester, dsep_test
    import numpy as np
    import random
    random.seed(129318290)
    np.random.seed(129318290)

    nnodes = 6
    nodes = set(range(nnodes))
    exp_nbrs = 2
    ngraphs = 100
    target_nodes = [random.randint(0, nnodes) for _ in range(ngraphs)]
    dags = cd.rand.directed_erdos(nnodes, exp_nbrs/(nnodes-1), ngraphs)
    ci_testers = [MemoizedCI_Tester(dsep_test, d) for d in dags]
    est_adjacent_list = [estimate_adjacent(nodes, target_node, ci_tester) for target_node, ci_tester in zip(target_nodes, ci_testers)]
    true_adjacent_list = [d.neighbors_of(target_node) for d, target_node in zip(dags, target_nodes)]
    match = np.array([est_adjacent == true_adjacent for est_adjacent, true_adjacent in zip(est_adjacent_list, true_adjacent_list)])
    print(np.where(~match))
    print([est_adjacent - true_adjacent for est_adjacent, true_adjacent in zip(est_adjacent_list, true_adjacent_list)])
    print([true_adjacent - true_adjacent for est_adjacent, true_adjacent in zip(est_adjacent_list, true_adjacent_list)])

    ix = 1
    print(dags[ix])
    print(target_nodes[ix])
    print(est_adjacent_list[ix])





from causaldag.classes import DAG
import numpy as np
from causaldag.utils import core_utils


class SampleDAG(DAG):
    def __init__(self, nodes, arcs):
        super().__init__(set(nodes), arcs)
        self.conditionals = dict()
        self._node_list = list(nodes)
        self._node2ix = core_utils.ix_map_from_list(self._node_list)

    def set_conditional(self, node, conditional_distribution):
        self.conditionals[node] = conditional_distribution

    def sample(self, nsamples: int = 1):
        samples = np.zeros((nsamples, len(self._nodes)))
        t = self.topological_sort()
        for sample_num in range(nsamples):
            for node in t:
                node_ix = self._node2ix[node]
                parent_ixs = [self._node2ix[p] for p in self._parents[node]]
                samples[sample_num, node_ix] = self.conditionals[node](samples[sample_num, parent_ixs])
        return samples

    def sample_interventional(self, interventions, nsamples: int = 1):
        pass



from .invariance_tester import PlainInvarianceTester, MemoizedInvarianceTester, InvarianceTester
from .hsic import hsic_invariance_test
from .kci import kci_invariance_test
from .gauss_invariance import gauss_invariance_test, gauss_invariance_suffstat
from .oracle import dsep_invariance_test



import itertools as itr
from numpy import abs


def ix_map_from_list(l):
    return {e: i for i, e in enumerate(l)}


def defdict2dict(defdict, keys):
    factory = defdict.default_factory
    d = {k: factory(v) for k, v in defdict.items()}
    for k in keys:
        if k not in d:
            d[k] = factory()
    return d


def powerset(s, r_min=0, r_max=None):
    if r_max is None: r_max = len(s)
    return itr.chain(*(itr.combinations(s, r) for r in range(r_min, r_max+1)))


def to_set(o):
    if not isinstance(o, set):
        try:
            return set(o)
        except TypeError:
            if o is None:
                return set()
            return {o}
    return o


def to_list(o):
    if not isinstance(o, list):
        try:
            return list(o)
        except TypeError:
            if o is None:
                return []
            return [o]
    return o


def is_symmetric(matrix, tol=1e-8):
    return (abs(matrix - matrix.T) <= tol).all()



from .dag import DAG, CycleError
from .pag import PAG
from .pdag import PDAG
from .gaussdag import GaussDAG
from .interventions import SoftIntervention, PerfectIntervention, SoftInterventionalDistribution, \
    PerfectInterventionalDistribution, ScalingIntervention, GaussIntervention, BinaryIntervention, \
    ConstantIntervention, MultinomialIntervention
from .ancestral_graph import AncestralGraph, SpouseError, AdjacentError, NeighborError
from . import ancestral_graph, dag, pdag
from .undirected_graph import UndirectedGraph
from .ggm import GGM
from .sample_dag import SampleDAG





from typing import Dict
from math import erf
import numba
from numpy import sqrt, log1p, abs, ix_, diag, corrcoef, errstate, cov
from numpy.linalg import inv
from . import MemoizedCI_Tester


@numba.jit
def numba_inv(A):
    return inv(A)


def gauss_ci_suffstat(samples, invert=True):
    """
    Helper function to compute the sufficient statistics for the gauss_ci_test from data.

    Parameters
    ----------
    samples:
        (n x p) matrix, where n is the number of samples and p is the number of variables.
    invert:
        if True, compute the inverse correlation matrix, and normalize it into the partial correlation matrix. This
        will generally speed up the gauss_ci_test if large conditioning sets are used.

    Return
    ------
    dictionary of sufficient statistics
    """
    n, p = samples.shape
    S = cov(samples, rowvar=False)
    C = corrcoef(samples, rowvar=False)
    if invert and n >= p:
        K = numba_inv(C)
        P = numba_inv(S)
        rho = K/sqrt(diag(K))/sqrt(diag(K))[:, None]
        return dict(P=P, S=S, C=C, n=n, K=K, rho=rho)
    return dict(S=S, C=C, n=n)


def compute_partial_correlation(suffstat, i, j, cond_set=None):
    """
    Compute the partial correlation between i and j given C

    Parameters
    ----------
    suffstat:
        dictionary containing:
        'n' -- number of samples
        'C' -- correlation matrix
        'K' (optional) -- inverse correlation matrix
        'rho' (optional) -- partial correlation matrix (K, normalized so diagonals are 1).
    i:
        position of first variable in correlation matrix.
    j:
        position of second variable in correlation matrix.
    cond_set:
        positions of conditioning set in correlation matrix.

    Return
    ------
    partial correlation
    """
    C = suffstat.get('C')
    p = C.shape[0]
    rho = suffstat.get('rho')
    K = suffstat.get('K')

    # === COMPUTE PARTIAL CORRELATION
    # partial correlation is correlation if there is no conditioning
    if cond_set is None or len(cond_set) == 0:
        r = C[i, j]
    # used closed-form
    elif len(cond_set) == 1:
        k = list(cond_set)[0]
        r = (C[i, j] - C[i, k]*C[j, k]) / sqrt((1 - C[j, k]**2) * (1 - C[i, k]**2))
    # when conditioning on everything, partial correlation comes from normalized precision matrix
    elif len(cond_set) == p - 2 and rho is not None:
        r = -rho[i, j]
    # faster to use Schur complement if conditioning set is large and precision matrix is pre-computed
    elif len(cond_set) >= p/2 and K is not None:
        rest = list(set(range(C.shape[0])) - {i, j, *cond_set})

        if len(rest) == 1:
            theta_ij = K[ix_([i, j], [i, j])] - K[ix_([i, j], rest)] @ K[ix_(rest, [i, j])] / K[rest[0], rest[0]]
        else:
            theta_ij = K[ix_([i, j], [i, j])] - K[ix_([i, j], rest)] @ numba_inv(K[ix_(rest, rest)]) @ K[ix_(rest, [i, j])]  # TODO: what to do if not invertible?
        r = -theta_ij[0, 1] / sqrt(theta_ij[0, 0] * theta_ij[1, 1])
    else:
        theta = numba_inv(C[ix_([i, j, *cond_set], [i, j, *cond_set])])  # TODO: what to do if not invertible?
        r = -theta[0, 1]/sqrt(theta[0, 0] * theta[1, 1])

    return r


def gauss_ci_test(suffstat: Dict, i, j, cond_set=None, alpha=None):
    """
    Test the null hypothesis that i and j are conditionally independent given cond_set via Fisher's z-transform.

    Parameters
    ----------
    suffstat:
        dictionary containing:
        'n' -- number of samples
        'C' -- correlation matrix
        'K' (optional) -- inverse correlation matrix
        'rho' (optional) -- partial correlation matrix (K, normalized so diagonals are 1).
    i:
        position of first variable in correlation matrix.
    j:
        position of second variable in correlation matrix.
    cond_set:
        positions of conditioning set in correlation matrix.
    alpha:
        Significance level.

    Return
    ------
    dictionary containing statistic, p_value, and reject.
    """
    n = suffstat['n']
    n_cond = 0 if cond_set is None else len(cond_set)
    alpha = 1/n if alpha is None else alpha

    r = compute_partial_correlation(suffstat, i, j, cond_set=cond_set)

    # === COMPUTE STATISTIC AND P-VALUE
    # note: log1p(2r/(1-r)) = log((1+r)/(1-r)) but is more numerically stable for r near 0
    # r = 1 causes warnings but gives the correct answer
    with errstate(divide='ignore', invalid='ignore'):
        statistic = sqrt(n - n_cond - 3) * abs(.5 * log1p(2*r/(1 - r)))
    # note: erf is much faster than norm.cdf
    p_value = 2*(1 - .5*(1 + erf(statistic/sqrt(2))))

    return dict(statistic=statistic, p_value=p_value, reject=p_value < alpha)


class MemoizedGaussCI_Tester(MemoizedCI_Tester):
    def __init__(self, suffstat: Dict, track_times=False, detailed=False, **kwargs):
        MemoizedCI_Tester.__init__(self, gauss_ci_test, suffstat, track_times=track_times, detailed=detailed)





# Author: Chandler Squires
"""
Base class for partially directed acyclic graphs
"""

from collections import defaultdict
from causaldag.utils import core_utils
import itertools as itr
import numpy as np
from typing import Set
from collections import namedtuple
from scipy.special import factorial
import networkx as nx
from typing import Set, FrozenSet

SmallDag = namedtuple('SmallDag', ['arcs', 'reversible_arcs', 'parents_dict', 'children_dict', 'level'])


class PDAG:
    def __init__(
            self,
            nodes: Set=set(),
            arcs: Set=set(),
            edges: Set=set(),
            known_arcs=set()
    ):
        self._nodes = set(nodes)
        self._arcs = set()
        self._edges = set()
        self._parents = defaultdict(set)
        self._children = defaultdict(set)
        self._neighbors = defaultdict(set)
        self._undirected_neighbors = defaultdict(set)
        for i, j in arcs:
            self._add_arc(i, j)
        for i, j in edges:
            self._add_edge(i, j)

        self._known_arcs = known_arcs.copy()

    @classmethod
    def from_amat(cls, amat):
        """Return a PDAG with arcs/edges given by amat
        """
        nrows, ncols = amat.shape
        arcs = set()
        edges = set()
        for (i, j), val in np.ndenumerate(amat):
            if val != 0:
                if (j, i) in arcs:
                    arcs.remove((j, i))
                    edges.add((i, j))
                else:
                    arcs.add((i, j))
        return PDAG(set(range(nrows)), arcs, edges)

    def to_nx(self):
        if self._arcs:
            raise NotImplementedError
        g = nx.Graph()
        g.add_edges_from(self._edges)
        return g

    @classmethod
    def from_nx(cls, nx_graph):
        return PDAG(nodes=nx_graph.nodes, edges=nx_graph.edges)

    def __eq__(self, other):
        same_nodes = self._nodes == other._nodes
        same_arcs = self._arcs == other._arcs
        same_edges = self._edges == other._edges

        return same_nodes and same_arcs and same_edges

    def __str__(self):
        substrings = []
        for node in self._nodes:
            parents = self._parents[node]
            nbrs = self._undirected_neighbors[node]
            parents_str = ','.join(map(str, parents)) if len(parents) != 0 else ''
            nbrs_str = ','.join(map(str, nbrs)) if len(nbrs) != 0 else ''

            if len(parents) == 0 and len(nbrs) == 0:
                substrings.append('[{node}]'.format(node=node))
            else:
                substrings.append('[{node}|{parents}:{nbrs}]'.format(node=node, parents=parents_str, nbrs=nbrs_str))
        return ''.join(substrings)

    def copy(self):
        """Return a copy of the graph
        """
        return PDAG(nodes=self._nodes, arcs=self._arcs, edges=self._edges, known_arcs=self._known_arcs)

    # === PROPERTIES
    @property
    def nodes(self):
        return set(self._nodes)

    @property
    def nnodes(self):
        return len(self._nodes)

    @property
    def num_arcs(self):
        return len(self._arcs)

    @property
    def num_edges(self):
        return len(self._edges)

    @property
    def num_adjacencies(self):
        return self.num_arcs + self.num_edges

    @property
    def arcs(self):
        return set(self._arcs)

    @property
    def edges(self):
        return set(self._edges)

    @property
    def parents(self):
        return core_utils.defdict2dict(self._parents, self._nodes)

    @property
    def children(self):
        return core_utils.defdict2dict(self._children, self._nodes)

    @property
    def neighbors(self):
        return core_utils.defdict2dict(self._neighbors, self._nodes)

    @property
    def undirected_neighbors(self):
        return core_utils.defdict2dict(self._undirected_neighbors, self._nodes)

    @property
    def skeleton(self):
        return {frozenset({i, j}) for i, j in self._arcs | self._edges}

    @property
    def dominated_nodes(self):
        dominated_nodes = set()
        for node in self._nodes:
            num_nbrs = self.undirected_degree_of(node)
            if num_nbrs == 0:
                dominated_nodes.add(node)
            elif num_nbrs == 1:
                max_nbrs_of_nbrs = max(len(self._undirected_neighbors[nbr]) for nbr in self._undirected_neighbors[node])
                if max_nbrs_of_nbrs > 1:
                    dominated_nodes.add(node)
        return dominated_nodes

    def clique_size(self):
        if len(self._arcs) == 0:
            g = self.to_nx()
            return nx.chordal_graph_treewidth(g) + 1
        else:
            return max(cc.clique_size() for cc in self.chain_components())

    def max_cliques(self):
        if len(self._arcs) == 0:
            g = self.to_nx()
            m_cliques = nx.chordal_graph_cliques(g)
            return {frozenset(c) for c in m_cliques}
        else:
            return set.union(*(cc.max_cliques() for cc in self.chain_components()))

    # === PROPERTIES W/ ARGUMENTS
    def indegree_of(self, node):
        return len(self._parents[node])

    def outdegree_of(self, node):
        return len(self._children[node])

    def undirected_degree_of(self, node):
        return len(self._undirected_neighbors[node])

    def total_degree_of(self, node):
        return len(self._neighbors[node])

    def parents_of(self, node):
        return set(self._parents[node])

    def children_of(self, node):
        return set(self._children[node])

    def neighbors_of(self, node):
        return set(self._neighbors[node])

    def undirected_neighbors_of(self, node):
        return set(self._undirected_neighbors[node])

    def has_edge(self, i, j):
        """Return True if the graph contains the edge i--j
        """
        return frozenset({i, j}) in self._edges

    def has_arc(self, i, j):
        """Return True if the graph contains the arc i->j"""
        return (i, j) in self._arcs

    def has_edge_or_arc(self, i, j):
        """Return True if the graph contains the edge i--j or an arc i->j or i<-j
        """
        return (i, j) in self._arcs or (j, i) in self._arcs or self.has_edge(i, j)

    def vstructs(self):
        vstructs = set()
        for node in self._nodes:
            for p1, p2 in itr.combinations(self._parents[node], 2):
                if p1 not in self._parents[p2] and p2 not in self._parents[p1]:
                    vstructs.add((p1, node))
                    vstructs.add((p2, node))
        return vstructs

    def _undirected_reachable(self, node, tmp, visited):
        visited.add(node)
        tmp.add(node)
        for nbr in filter(lambda nbr: nbr not in visited, self._undirected_neighbors[node]):
            tmp = self._undirected_reachable(nbr, tmp, visited)
        return tmp

    def chain_components(self, rename=False):
        """Return the chain components of this graph.

        Return
        ------
        List[Set[node]]
            Return the partition of nodes coming from the relation of reachability by undirected edges.
        """
        node_queue = self._nodes.copy()
        components = []
        visited_nodes = set()

        while node_queue:
            node = node_queue.pop()
            if node not in visited_nodes:
                reachable = self._undirected_reachable(node, set(), visited_nodes)
                if len(reachable) > 1: components.append(reachable)

        return [self.induced_subgraph(c, rename=rename) for c in components]

    def induced_subgraph(self, nodes, rename=False):
        if rename:
            ixs = dict(map(reversed, enumerate(nodes)))
            new_nodes = set(range(len(nodes)))
            arcs = {(ixs[i], ixs[j]) for i, j in self._arcs if i in nodes and j in nodes}
            edges = {(ixs[i], ixs[j]) for i, j in self._edges if i in nodes and j in nodes}
        else:
            new_nodes = nodes
            arcs = {(i, j) for i, j in self._arcs if i in nodes and j in nodes}
            edges = {(i, j) for i, j in self._edges if i in nodes and j in nodes}
        return PDAG(nodes=new_nodes, arcs=arcs, edges=edges)

    def interventional_cpdag(self, dag, intervened_nodes):
        cut_edges = set()
        for node in intervened_nodes:
            cut_edges.update(dag.incident_arcs(node))
        p = PDAG(self._nodes, self._arcs | cut_edges, self._edges - {frozenset({i, j}) for i, j in cut_edges})
        p.to_complete_pdag()
        return p

    # === MUTATORS
    def _add_arc(self, i, j):
        self._nodes.add(i)
        self._nodes.add(j)
        self._arcs.add((i, j))

        self._neighbors[i].add(j)
        self._neighbors[j].add(i)

        self._children[i].add(j)
        self._parents[j].add(i)

    def _add_edge(self, i, j):
        self._nodes.add(i)
        self._nodes.add(j)
        self._edges.add(frozenset({i, j}))

        self._neighbors[i].add(j)
        self._neighbors[j].add(i)

        self._undirected_neighbors[i].add(j)
        self._undirected_neighbors[j].add(i)

    def remove_edge(self, i, j, ignore_error=False):
        try:
            self._edges.remove(frozenset({i, j}))
            self._neighbors[i].remove(j)
            self._neighbors[j].remove(i)
            self._undirected_neighbors[i].remove(j)
            self._undirected_neighbors[j].remove(i)
        except KeyError as e:
            if ignore_error:
                pass
            else:
                raise e

    def remove_node(self, node):
        """Remove a node from the graph
        """
        self._nodes.remove(node)
        self._arcs = {(i, j) for i, j in self._arcs if i != node and j != node}
        self._edges = {frozenset({i, j}) for i, j in self._edges if i != node and j != node}
        for child in self._children[node]:
            self._parents[child].remove(node)
            self._neighbors[child].remove(node)
        for parent in self._parents[node]:
            self._children[parent].remove(node)
            self._neighbors[parent].remove(node)
        for u_nbr in self._undirected_neighbors[node]:
            self._undirected_neighbors[u_nbr].remove(node)
            self._neighbors[u_nbr].remove(node)
        del self._parents[node]
        del self._children[node]
        del self._neighbors[node]
        del self._undirected_neighbors[node]

    def replace_edge_with_arc(self, arc, ignore_error=False):
        try:
            self._replace_edge_with_arc(arc)
        except KeyError as e:
            if ignore_error:
                pass
            else:
                raise e

    def _replace_arc_with_edge(self, arc):
        self._arcs.remove(arc)
        self._edges.add(frozenset({*arc}))
        i, j = arc
        self._parents[j].remove(i)
        self._children[i].remove(j)
        self._undirected_neighbors[i].add(j)
        self._undirected_neighbors[j].add(i)

    def _replace_edge_with_arc(self, arc):
        self._edges.remove(frozenset({*arc}))
        self._arcs.add(arc)
        i, j = arc
        self._parents[j].add(i)
        self._children[i].add(j)
        self._undirected_neighbors[i].remove(j)
        self._undirected_neighbors[j].remove(i)

    def assign_parents(self, node, parents, verbose=False):
        for p in parents:
            self._replace_edge_with_arc((p, node))
        for c in self._undirected_neighbors[node] - parents:
            self._replace_edge_with_arc((node, c))
        self.to_complete_pdag(verbose=verbose)

    def to_complete_pdag(self, verbose=False, solve_conflict=False):
        """
        Replace with arcs those edges whose orientations can be determined by Meek rules:
        =====
        See Koller & Friedman, Algorithm 3.5

        """
        if solve_conflict:
            raise NotImplementedError

        PROTECTED = 'P'  # indicates that some configuration definitely exists to protect the edge
        UNDECIDED = 'U'  # indicates that some configuration exists that could protect the edge
        NOT_PROTECTED = 'N'  # indicates no possible configuration that could protect the edge

        edges1 = {(i, j) for i, j in self._edges}
        undecided_arcs = edges1 | {(j, i) for i, j in edges1}
        arc_flags = {arc: PROTECTED for arc in self._arcs}
        arc_flags.update({arc: UNDECIDED for arc in undecided_arcs})

        while undecided_arcs:
            for arc in undecided_arcs:
                i, j = arc
                flag = NOT_PROTECTED

                # check configuration (a) -- causal chain
                s = ''
                for k in self._parents[i]:
                    if not self.has_edge_or_arc(k, j):
                        if arc_flags[(k, i)] == PROTECTED:
                            flag = PROTECTED
                            s = f': {k}->{i}-{j}'
                            break
                        else:
                            flag = UNDECIDED
                if verbose: print(f'{arc} marked {flag} by (a){s}')

                # check configuration (b) -- acyclicity
                s = ''
                if flag != PROTECTED:
                    for k in self._parents[j]:
                        if i in self._parents[k]:
                            if arc_flags[(i, k)] == PROTECTED and arc_flags[(k, j)] == PROTECTED:
                                flag = PROTECTED
                                s = f': {k}->{j}-{i}->{k}'
                                break
                            else:
                                flag = UNDECIDED
                    if verbose: print(f'{arc} marked {flag} by (b){s}')

                # check configuration (d)
                s = ''
                if flag != PROTECTED:
                    for k1, k2 in itr.combinations(self._parents[j], 2):
                        if self.has_edge(i, k1) and self.has_edge(i, k2) and not self.has_edge_or_arc(k1, k2):
                            if arc_flags[(k1, j)] == PROTECTED and arc_flags[(k2, j)] == PROTECTED:
                                flag = PROTECTED
                                s = f': {i}-{k1}->{j}<-{k2}-{i}'
                                break
                            else:
                                flag = UNDECIDED
                    if verbose: print(f'{arc} marked {flag} by (c){s}')

                arc_flags[arc] = flag

            if all(arc_flags[arc] == NOT_PROTECTED for arc in undecided_arcs): break

            for arc in undecided_arcs.copy():
                if arc_flags[arc] == PROTECTED:
                    if not solve_conflict:
                        if self.has_arc(arc[1], arc[0]):  # arc has already been oriented the opposite way
                            continue
                    undecided_arcs.remove(arc)
                    undecided_arcs.remove((arc[1], arc[0]))
                    self._replace_edge_with_arc(arc)

    def remove_unprotected_orientations(self, verbose=False):
        """
        Replace with edges those arcs whose orientations cannot be determined by either:
        - prior knowledge, or
        - Meek rules

        =====
        See Koller & Friedman, Algorithm 3.5

        """
        PROTECTED = 'P'  # indicates that some configuration definitely exists to protect the edge
        UNDECIDED = 'U'  # indicates that some configuration exists that could protect the edge
        NOT_PROTECTED = 'N'  # indicates no possible configuration that could protect the edge

        undecided_arcs = self._arcs - self._known_arcs
        arc_flags = {arc: PROTECTED for arc in self._known_arcs}
        arc_flags.update({arc: UNDECIDED for arc in undecided_arcs})

        while undecided_arcs:
            for arc in undecided_arcs:
                i, j = arc
                flag = NOT_PROTECTED

                # check configuration (a) -- causal chain
                for k in self._parents[i]:
                    if not self.has_edge_or_arc(k, j):
                        if arc_flags[(k, i)] == PROTECTED:
                            flag = PROTECTED
                            break
                        else:
                            flag = UNDECIDED
                        if verbose: print('{edge} marked {flag} by (a)'.format(edge=arc, flag=flag))

                # check configuration (b) -- acyclicity
                if flag != PROTECTED:
                    for k in self._parents[j]:
                        if i in self._parents[k]:
                            if arc_flags[(i, k)] == PROTECTED and arc_flags[(k, j)] == PROTECTED:
                                flag = PROTECTED
                                break
                            else:
                                flag = UNDECIDED
                            if verbose: print('{edge} marked {flag} by (b)'.format(edge=arc, flag=flag))

                # check configuration (d)
                if flag != PROTECTED:
                    for k1, k2 in itr.combinations(self._parents[j], 2):
                        if self.has_edge(i, k1) and self.has_edge(i, k2) and not self.has_edge_or_arc(k1, k2):
                            if arc_flags[(k1, j)] == PROTECTED and arc_flags[(k2, j)] == PROTECTED:
                                flag = PROTECTED
                            else:
                                flag = UNDECIDED
                            if verbose: print('{edge} marked {flag} by (c)'.format(edge=arc, flag=flag))

                arc_flags[arc] = flag

            for arc in undecided_arcs.copy():
                if arc_flags[arc] != UNDECIDED:
                    undecided_arcs.remove(arc)
                if arc_flags[arc] == NOT_PROTECTED:
                    self._replace_arc_with_edge(arc)

    def add_known_arc(self, i, j):
        if (i, j) in self._known_arcs:
            return
        self._known_arcs.add((i, j))
        self._edges.remove(frozenset({i, j}))
        self.remove_unprotected_orientations()

    def add_known_arcs(self, arcs):
        raise NotImplementedError

    # === MUTATORS
    def to_amat(self, node_list=None, mode='dataframe'):
        """Return an adjacency matrix for the graph
        """
        if node_list is None:
            node_list = sorted(self._nodes)
        node2ix = {node: i for i, node in enumerate(node_list)}

        shape = (len(self._nodes), len(self._nodes))
        if mode == 'dataframe' or mode == 'numpy':
            amat = np.zeros(shape, dtype=int)
        else:
            from scipy.sparse import lil_matrix
            amat = lil_matrix(shape, dtype=int)

        for source, target in self._arcs:
            amat[node2ix[source], node2ix[target]] = 1
        for i, j in self._edges:
            amat[node2ix[i], node2ix[j]] = 1
            amat[node2ix[j], node2ix[i]] = 1

        if mode == 'dataframe':
            from pandas import DataFrame
            return DataFrame(amat, index=node_list, columns=node_list)
        else:
            return amat, node_list

    def _possible_sinks(self):
        return {node for node in self._nodes if len(self._children[node]) == 0}

    def _neighbors_covered(self, node):
        return {node2: self.neighbors[node2] - {node} == self.neighbors[node] for node2 in self._nodes}

    def to_dag(self):
        """
        Return a DAG that is consistent with this CPDAG.

        Returns
        -------
        d

        Examples
        --------
        TODO
        """
        from causaldag import DAG

        pdag2 = self.copy()
        arcs = set()
        while len(pdag2._edges) + len(pdag2._arcs) != 0:
            is_sink = lambda n: len(pdag2._children[n]) == 0
            no_vstructs = lambda n: all(
                (pdag2._neighbors[n] - {u_nbr}).issubset(pdag2._neighbors[u_nbr])
                for u_nbr in pdag2._undirected_neighbors[n]
            )
            sink = next((n for n in pdag2._nodes if is_sink(n) and no_vstructs(n)), None)
            if sink is None:
                break
            arcs.update((nbr, sink) for nbr in pdag2._neighbors[sink])
            pdag2.remove_node(sink)

        return DAG(arcs=arcs)

    # === MEC
    def mec_size(self):
        """Return the number of DAGs in the MEC represented by this PDAG
        """
        raise NotImplementedError

        if self.num_arcs != 0:
            pass
        elif self.num_edges == nnodes - 1:
            return nnodes
        elif self.num_edges == nnodes*(nnodes-1)/2:
            return factorial(nnodes)
        else:
            pass

    def exact_sample(self, save_sampler=True, nsamples=1):
        """Return a DAG sampled uniformly at random from the MEC represented by this PDAG
        """
        raise NotImplementedError

    def all_dags(self, verbose=False):
        """Return all DAGs consistent with this PDAG
        """
        dag = self.to_dag()
        arcs = dag._arcs
        all_arcs = set()

        orig_reversible_arcs = dag.reversible_arcs() - self._known_arcs
        orig_parents_dict = dag.parents
        orig_children_dict = dag.children

        level = 0
        q = [SmallDag(arcs, orig_reversible_arcs, orig_parents_dict, orig_children_dict, level)]
        while q:
            dag = q.pop()
            all_arcs.add(frozenset(dag.arcs))
            for i, j in dag.reversible_arcs:
                new_arcs = frozenset({arc for arc in dag.arcs if arc != (i, j)} | {(j, i)})
                if new_arcs not in all_arcs:
                    new_parents_dict = {}
                    new_children_dict = {}
                    for node in dag.parents_dict.keys():
                        parents = set(dag.parents_dict[node])
                        children = set(dag.children_dict[node])
                        if node == i:
                            new_parents_dict[node] = parents | {j}
                            new_children_dict[node] = children - {j}
                        elif node == j:
                            new_parents_dict[node] = parents - {i}
                            new_children_dict[node] = children | {i}
                        else:
                            new_parents_dict[node] = parents
                            new_children_dict[node] = children

                    new_reversible_arcs = dag.reversible_arcs.copy()
                    for k in dag.parents_dict[j]:
                        if (new_parents_dict[j] - {k}) == new_parents_dict[k] and (k, j) not in self._known_arcs:
                            new_reversible_arcs.add((k, j))
                        else:
                            new_reversible_arcs.discard((k, j))
                    for k in dag.children_dict[j]:
                        if new_parents_dict[j] == (new_parents_dict[k] - {j}) and (j, k) not in self._known_arcs:
                            new_reversible_arcs.add((j, k))
                        else:
                            new_reversible_arcs.discard((j, k))
                    for k in dag.parents_dict[i]:
                        if (new_parents_dict[i] - {k}) == new_parents_dict[k] and (k, i) not in self._known_arcs:
                            new_reversible_arcs.add((k, i))
                        else:
                            new_reversible_arcs.discard((k, i))
                    for k in dag.children_dict[i]:
                        if new_parents_dict[i] == (new_parents_dict[k] - {i}) and (i, k) not in self._known_arcs:
                            new_reversible_arcs.add((i, k))
                        else:
                            new_reversible_arcs.discard((i, k))

                    q.append(SmallDag(new_arcs, new_reversible_arcs, new_parents_dict, new_children_dict, dag.level+1))

        return all_arcs

    # === COMPARISON
    def shd(self, other):
        """Return the structural Hamming distance between this PDAG and another
        """
        self_undirected = {frozenset({*arc}) for arc in self._arcs} | self._edges
        other_undirected = {frozenset({*arc}) for arc in other._arcs} | other._edges
        num_additions = len(self_undirected - other_undirected)
        num_deletions = len(other_undirected - self_undirected)
        diff_type = {
            (i, j) for i, j in self_undirected & other_undirected
            if ((i, j) in self._arcs and (i, j) not in other._arcs) or
               ((j, i) in self._arcs and (j, i) not in other._arcs) or
               ((i, j) in self._edges and (i, j) not in other._edges)
        }
        return num_additions + num_deletions + len(diff_type)

    def shd_skeleton(self, other) -> int:
        return len(self.skeleton.symmetric_difference(other.skeleton))


if __name__ == '__main__':
    from causaldag.rand import directed_erdos

    g = directed_erdos(10, .5)
    c = g.cpdag()
    a1 = c.to_amat()
    a2, _ = c.to_amat(mode='numpy')
    a3, _ = c.to_amat(mode='sparse')



import pygam
import numpy as np


def residuals(samples, i, j, cond_set):
    cond_set = list(cond_set)
    g = pygam.GAM()
    g.fit(samples[:, cond_set], samples[:, i])
    residuals_i = g.deviance_residuals(samples[:, cond_set], samples[:, i])
    g.fit(samples[:, cond_set], samples[:, j])
    residuals_j = g.deviance_residuals(samples[:, cond_set], samples[:, j])

    return residuals_i, residuals_j







"""
===============================
Estimation of differences between undirected graphs.
===============================
This module contains functions for computing the difference undirected graph given two data sets.
References
----------
.. [1] Liu, S., Quinn, J. A., Gutmann, M. U., Suzuki, T., & Sugiyama, M. (2014). 
   Direct learning of sparse changes in Markov networks by density ratio estimation. 
   Neural computation, 26(6), 1169-1197.
   [2] http://allmodelsarewrong.net/kliep_sparse/demo_sparse.html
"""


import numpy as np 
import scipy
import networkx as nx 


def dci_undirected_graph(X1, X2, alpha=1.0, max_iter=1000, edge_threshold=0.05):
    """
    Estimates the difference between two undirected graphs directly from two data sets
    using Kullback-Leibler importance estimation procedure (KLIEP).

    Parameters
    ----------
    X1: array, shape = [n_samples, n_features]
        First dataset.    
    X2: array, shape = [n_samples, n_features]
        Second dataset.
    alpha: float, default = 1.0
        L1 regularization parameter.
    max_iter: int, default = 1000
        Maximum number of iterations for gradient descent.
    edge_threshold: float, default = 0.05
        Edge weight cutoff for keeping an edge (all edges above or equal to this threshold are kept).

    Returns
    -------
    difference_ug: list
        List of tuple of edges in the difference undirected graph.
    difference_nodes: set
        Nodes in the difference undirected graph.    
    """

    k1 = kernel_linear(X1)
    k2 = kernel_linear(X2)
    theta = naive_subgradient_descent(k1, k2, alpha=alpha, max_iter=1000)
    difference_ug, difference_nodes = compute_difference_graph(X1, theta, edge_threshold=edge_threshold)
    return difference_ug, difference_nodes


def kernel_linear(X):
    """
    Computes polynomial features (order = 2) based on data.
    """
    n, d = X.shape
    kernel_features = np.zeros((n, int((d*(d-1))/2)))
    for i in range(n):
        t = np.matmul(X[i,:].reshape(d,1), X[i,:].reshape(d,1).T)
        kernel_features[i,:] = t[np.triu_indices_from(t,1)]

    kernel_features = np.concatenate((kernel_features, X**2), axis=1)
    return kernel_features


def llkliep(theta, k1, k2):
    """
    Computes the log-likelihood of the model and the gradient.
    """
    loglik = -np.mean(np.matmul(theta.T, k1.T), 1) + scipy.special.logsumexp(np.matmul(theta.T, k2.T), 1)
    log_g_q = np.matmul(theta.T, k2.T) - scipy.special.logsumexp(np.matmul(theta.T, k2.T), 1)
    g_q = np.exp(log_g_q)
    grad = -np.mean(k1, 0).reshape((-1, 1)) + np.matmul(k2.T, g_q.T)
    return loglik[0], grad


def naive_subgradient_descent(k1, k2, alpha=1, max_iter=1000):
    """
    Performs gradient updates to find parameters that maximize the log-likelihood.

    Parameters
    ----------
    k1: array, shape = [n_samples, n_features]
        First dataset after featurization.    
    k2: array, shape = [n_samples, n_features]
        Second dataset after featurization.
    alpha: float, default = 1.0
        L1 regularization parameter.
    max_iter: int, default = 1000
        Maximum number of iterations for gradient descent.

    Returns
    -------
    theta: array
        Estimated parameters corresponding to the difference undirected graph.  
    """

    # initialize variables
    d = k1.shape[1]
    theta = np.zeros((d, 1))
    step = 1
    slength = np.inf
    iter = 0
    loglik_old = 1e20

    while (slength > 1e-5) & (iter < max_iter):
        loglik, grad = llkliep(theta, k1, k2)
        g = np.zeros(grad.shape)

        ids = theta.nonzero()
        g[ids] = grad[ids] + alpha*np.sign(theta[ids])

        zero_ids = theta == 0
        ids = zero_ids & (grad > alpha)
        g[ids] = grad[ids] - alpha

        ids = zero_ids & (grad < -alpha)
        g[ids] = gt[ids] + alpha

        # update theta parameters
        theta = theta - step*g/(iter + 1)
        slength = step*np.linalg.norm(g)/(iter + 1)

        loglik_diff = np.abs(loglik - loglik_old)
        iter = iter + 1
        if iter == max_iter:
            print('Maximum iteration reached')
    return theta


def compute_difference_graph(X, theta, edge_threshold=0.05):
    """
    Obtain difference undirected graph and corresponding nodes from estimated parameters.
    """
    n, d = x.shape
    delta_ug = np.zeros((d,d))
    delta_ug[np.triu_indices(d, 1)] = theta[0:-d].flatten()
    delta_ug = delta_ug + delta_ug.T
    # remove edges that are below cutoff threshold
    delta_ug[np.abs(delta_ug) < edge_threshold] = 0
    g = nx.from_numpy_matrix(delta_ug)
    return list(g.edges()), set(np.unique(g.edges()))


from causaldag import DAG


cancer_network = DAG(arcs={
    ('Pollution', 'Cancer'),
    ('Smoker', 'Cancer'),
    ('Cancer', 'Xmy'),
    ('Cancer', 'Dysponoea')
})

earthquake_network = DAG(arcs={
    ('Burglary', 'Alarm'),
    ('Earthquake', 'Alarm'),
    ('Alarm', 'JohnCalls'),
    ('Alarm', 'MaryCalls')
})

sachs_network = DAG(arcs={
    ('PKC', 'PKA'),
    ('PKC', 'Jnk'),
    ('PKC', 'P38'),
    ('PKC', 'Raf'),
    ('PKC', 'Mek'),
    ('PKA', 'Jnk'),
    ('PKA', 'P38'),
    ('PKA', 'Raf'),
    ('PKA', 'Mek'),
    ('PKA', 'Erk'),
    ('PKA', 'Akt'),
    ('Raf', 'Mek'),
    ('Mek', 'Erk'),
    ('Erk', 'Akt'),
    ('Plcg', 'PIP3'),
    ('Plcg', 'PIP2'),
    ('PIP3', 'PIP2'),
})



import numpy as np
import numba

@numba.jit
def numba_inv(A):
    return np.linalg.inv(A)


def local_score(self, node, parents, suffstat, lambda_=None):
    lambda_ if lambda_ is not None else .5 * np.log2(self.n)
    C, i, p = suffstat['C'], node, parents

    var = C[i, i] if not p else C[i, i] - C[i, p] @ numba_inv(C[np.ix_(p, p)]) @ C[p, i]
    log_prob = -.5*self.n*(1 + np.log2(var/self.n))
    penalty_term = self.lambda_*len(1 + parents)

    return log_prob - penalty_term




from causaldag.utils.core_utils import powerset
from causaldag.utils.regression import RegressionHelper
from scipy.special import ncfdtr
from numpy.linalg import inv


def dci(
        nodes: set,
        difference_ug: set,
        suffstat1: dict,
        suffstat2: dict,
        alpha_skeleton: float=.01,
        alpha_orient: float=.01,
        max_set_size: int = None
):
    """
    Use the Difference Causal Inference (DCI) algorithm to estimate the difference-DAG between two settings.

    Parameters
    ----------
    nodes:
        Labels of nodes in the graph.
    suffstat1:
        Dictionary of sufficient statistics for the first dataset.
    suffstat2:
        Dictionary of sufficient statistics for the second dataset.
    alpha_skeleton:
        todo
    alpha_orient:
        todo
    difference_ug:
        Estimated set of edges in the difference UG.
    max_set_size:
        Maximum conditioning set size used to test regression invariance.

    See Also
    --------
    dci_skeleton, dci_orient

    Returns
    -------

    """
    rh1 = RegressionHelper(suffstat1)
    rh2 = RegressionHelper(suffstat2)
    skeleton = dci_skeleton(nodes, difference_ug, max_set_size=max_set_size, rh1=rh1, rh2=rh2, alpha=alpha_skeleton)
    return dci_orient(skeleton, max_set_size=max_set_size, alpha=alpha_orient, rh1=rh1, rh2=rh2)


def dci_skeleton(
        nodes: set,
        difference_ug: set,
        suffstat1: dict=None,
        suffstat2: dict=None,
        alpha: float=.05,
        max_set_size: int=None,
        rh1: RegressionHelper=None,
        rh2: RegressionHelper=None,
):
    """
    Perform Phase I of the Difference Causal Inference (DCI) algorithm, i.e., estimate the skeleton of the
    difference DAG.

    Parameters
    ----------
    nodes:
        Labels of nodes in the graph.
    difference_ug:
        Estimated set of edges in the difference UG.
    suffstat1:
        Dictionary of sufficient statistics for the first dataset.
    suffstat2:
        Dictionary of sufficient statistics for the second dataset.
    alpha:
        todo
    max_set_size:
        Maximum conditioning set size used to test regression invariance.
    rh1:
        todo
    rh2:
        todo

    See Also
    --------
    dci, dci_orient

    Returns
    -------

    """
    if rh1 is None:
        rh1 = RegressionHelper(suffstat1)
        rh2 = RegressionHelper(suffstat2)

    n1 = rh1.suffstat['n']
    n2 = rh2.suffstat['n']
    skeleton = {frozenset({i, j}) for i, j in difference_ug}
    for i, j in difference_ug:
        for cond_set in powerset(nodes - {i, j}, r_max=max_set_size):
            cond_set_i, cond_set_j = [*cond_set, j], [*cond_set, i]

            # calculate regression coefficients (j regressed on cond_set_j) for both datasets
            beta1_i, var1_i, precision1 = rh1.regression(i, cond_set_i)
            beta2_i, var2_i, precision2 = rh2.regression(i, cond_set_i)

            # compute statistic and p-value
            j_ix = cond_set_i.index(j)
            stat_i = (beta1_i[j_ix] - beta2_i[j_ix])**2 * inv(var1_i*precision1/(n1-1) + var2_i*precision2/(n2-1))[j_ix, j_ix]
            pval_i = ncfdtr(1, n1 + n2 - len(cond_set_i) - len(cond_set_j), 0, stat_i)
            pval_i = 2 * min(pval_i, 1 - pval_i)

            #  remove i-j from skeleton if i regressed on (j, cond_set) is invariant
            i_invariant = pval_i > alpha
            if i_invariant:
                skeleton.remove(frozenset({i, j}))
                break

            # calculate regression coefficients (i regressed on cond_set_i) for both datasets
            beta1_j, var1_j, precision1 = rh1.regression(j, cond_set_j)
            beta2_j, var2_j, precision2 = rh2.regression(j, cond_set_j)

            # compute statistic and p-value
            i_ix = cond_set_j.index(i)
            stat_j = (beta1_j[i_ix] - beta2_j[i_ix]) ** 2 * inv(var1_j*precision1/(n1-1) + var2_j*precision2/(n2-1))[i_ix, i_ix]
            pval_j = 1 - ncfdtr(1, n1 + n2 - len(cond_set_i) - len(cond_set_j), 0, stat_j)
            pval_j = 2 * min(pval_j, 1 - pval_j)

            #  remove i-j from skeleton if j regressed on (i, cond_set) is invariant
            j_invariant = pval_j > alpha
            if j_invariant:
                skeleton.remove(frozenset({i, j}))
                break

    return skeleton


def dci_orient(
        skeleton: set,
        suffstat1: dict = None,
        suffstat2: dict = None,
        alpha: float = .05,
        max_set_size: int = None,
        rh1: RegressionHelper=None,
        rh2: RegressionHelper=None,
):
    """
    Perform Phase I of the Difference Causal Inference (DCI) algorithm, i.e., orient edges in the skeleton of the
    difference DAG.

    Parameters
    ----------
    skeleton:
        The estimated skeleton of the difference DAG.
    suffstat1:
        Dictionary of sufficient statistics for the first dataset.
    suffstat2:
        Dictionary of sufficient statistics for the second dataset.
    alpha:
        todo
    max_set_size:
        Maximum conditioning set size used to test regression invariance.
    rh1:
        todo
    rh2:
        todo

    See Also
    --------
    dci, dci_skeleton

    Returns
    -------

    """
    if rh1 is None:
        rh1 = RegressionHelper(suffstat1)
        rh2 = RegressionHelper(suffstat2)

    nodes = {i for i, j in skeleton} | {j for i, j in skeleton}
    oriented_edges = set()

    n1 = rh1.suffstat['n']
    n2 = rh2.suffstat['n']
    for i, j in skeleton:
        for cond_i, cond_j in zip(powerset(nodes - {i}, r_max=max_set_size), powerset(nodes - {j}, r_max=max_set_size)):
            # compute residual variances for i
            beta1_i, var1_i, _ = rh1.regression(i, cond_i)
            beta2_i, var2_i, _ = rh2.regression(i, cond_i)
            # check pvalue
            pvalue_i = ncfdtr(n1 - len(cond_i), n2 - len(cond_i), 0, var1_i/var2_i)
            pvalue_i = min(pvalue_i, 1-pvalue_i)
            if pvalue_i > alpha:
                oriented_edges.add((j, i) if j in cond_i else (i, j))
                break

            # compute residual variances for j
            beta1_j, var1_j, _ = rh1.regression(i, cond_j)
            beta2_j, var2_j, _ = rh2.regression(i, cond_j)
            # check pvalue
            pvalue_j = ncfdtr(n1 - len(cond_j), n2 - len(cond_j), 0, var1_j / var2_j)
            pvalue_j = min(pvalue_j, 1 - pvalue_j)
            if pvalue_j > alpha:
                oriented_edges.add((i, j) if i in cond_j else (j, i))
                break

    unoriented_edges = skeleton - {frozenset({i, j}) for i, j in oriented_edges}
    return oriented_edges, unoriented_edges
    

if __name__ == '__main__':
    from causaldag.rand.graphs import directed_erdos, rand_weights
    import causaldag as cd
    from causaldag.utils.ci_tests import gauss_ci_suffstat
    import itertools as itr

    import sys
    import os
    sys.path.append(os.path.join(os.path.expanduser('~'), 'dropbox', 'learning differences of DAGs', 'py_work'))
    from algs.skeleton.yc_skeleton import infer_skeleton
    from algs.orienting.yc_orienting import infer_directions

    nnodes = 10
    nodes_ = set(range(nnodes))
    exp_nbrs = 2
    nsamples1 = 10000
    nsamples2 = 10100
    alpha_ = .05
    candidate_edges = set(itr.combinations(nodes_, 2))

    d = directed_erdos(nnodes, exp_nbrs/(nnodes-1))
    g1 = rand_weights(d)
    amat1 = g1.to_amat()
    amat2 = amat1.copy()
    amat2[0, 1] += 10
    g2 = cd.GaussDAG.from_amat(amat2)

    samples1 = g1.sample(nsamples1)
    samples2 = g2.sample(nsamples2)
    suff1 = gauss_ci_suffstat(samples1)
    suff2 = gauss_ci_suffstat(samples2)

    # skel = dci_skeleton(nodes_, candidate_edges, suff1, suff2, alpha=alpha_)
    # print(len(skel))

    # skel2, _, _, _ = infer_skeleton(samples1, samples2, candidate_edges, changed_nodes=nodes_, alpha=alpha_)
    skel = {frozenset({0, 1})}
    oriented_edges_, unoriented_edges_ = dci_orient(skel, suff1, suff2)
    oriented_edges2 = infer_directions(samples1, samples2, skel, changed_nodes=nodes_, alpha=alpha_)
    print(oriented_edges_, unoriented_edges_)
    print(oriented_edges2)






from functools import wraps
import time


def timed(func):
    """This decorator prints the execution time for the decorated function."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print("{} ran in {}s".format(func.__name__, round(end - start, 2)))
        return result
    return wrapper



from .ci_tester import MemoizedCI_Tester, PlainCI_Tester, CI_Tester, PooledCI_Tester
from .gauss_ci import gauss_ci_test, gauss_ci_suffstat, compute_partial_correlation
from .hsic import hsic_test_vector, hsic_test
from .kci import kci_test_vector, ki_test_vector, kci_test
from .oracle import dsep_test, msep_test




from numpy import ix_
from numpy.linalg import inv


class RegressionHelper:
    def __init__(self, suffstat):
        self.suffstat = suffstat
        self.S = suffstat['S']
        self.P = suffstat['P']
        self.p = self.suffstat['S'].shape[0]
        self.n = self.suffstat['n']

    def regression(self, i: int, c: list=None):
        S, P = self.S, self.P
        d = [j for j in range(self.p) if j not in c and j != i] + [i]

        if c is None or len(c) == 0:
            coefs = []
            var = S[i, i]
            S_inv = None

        # use Schur complement when conditioning to keep inverted submatrix small
        elif len(c) < self.p / 2:
            S_inv = inv(S[ix_(c, c)])
            coefs = S_inv @ S[c, i]
            var = S[i, i] - S[i, c] @ S_inv @ S[c, i]

        # use Schur complement when marginalizing to keep inverted submatrix small
        else:
            P_inv = inv(P[ix_(d, d)])
            S_inv = P[ix_(c, c)] - P[ix_(c, d)] @ P_inv @ P[ix_(d, c)]
            coefs = S_inv @ S[c, i]
            var = inv(P[ix_(d, d)])[-1, -1]

        # correct the variance to account for the number of degrees of freedom
        var = var * self.n / (self.n - len(c) + 1)

        return coefs, var, S_inv


if __name__ == '__main__':
    from causaldag.rand.graphs import directed_erdos, rand_weights
    from causaldag.utils.ci_tests import gauss_ci_suffstat
    from causaldag.utils.core_utils import powerset
    from sklearn.linear_model import LinearRegression
    import numpy as np
    from tqdm import trange

    nnodes = 10
    nodes = set(range(nnodes))
    exp_nbrs = 2
    nsamples = 100
    dag = directed_erdos(nnodes, exp_nbrs/(nnodes-1))
    gdag = rand_weights(dag)
    samples = gdag.sample(nsamples)
    suff = gauss_ci_suffstat(samples)
    reg_helper = RegressionHelper(suff)
    lr = LinearRegression()

    for i in trange(nnodes):
        for c in powerset(nodes - {i}, r_min=1):
            c = list(c)
            coefs, var = reg_helper.regression(i, c)
            lr.fit(samples[:, c], samples[:, i])
            var2 = np.var(samples[:, c] @ coefs - samples[:, i], ddof=len(c))
            if not np.isclose(coefs, lr.coef_).all():
                print(coefs, lr.coef_)
            if not np.isclose(var, var2):
                print(var, var2)




import setuptools

setuptools.setup(
    name='causaldag',
    version='0.1a.127',
    description='Causal DAG manipulation and inference',
    long_description='CausalDAG is a Python package for the creation, manipulation, and learning of Causal DAGs.',
    author='Chandler Squires',
    author_email='chandlersquires18@gmail.com',
    packages=setuptools.find_packages(exclude=['tests']),
    python_requires='>3.5.0',
    zip_safe=False,
    classifiers=[
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    install_requires=[
        'scipy',
        'dataclasses',
        'numpy',
        # 'scikit_sparse',
        'matplotlib',
        'pandas',
        'scikit_learn',
        'typing',
        'pygam',
        'numba',
        'networkx',
    ]
)

